TRAINABLE_MODEL_TYPES = [
    'visual-classifier', 'visual-detector', 'visual-segmenter', 'visual-embedder', 'clusterer',
    'text-classifier', 'embedding-classifier', 'text-to-text'
]
MAX_MODEL_PREDICT_INPUTS = 128
